#!/usr/bin/env python

import os
import sys
import gzip
import logging
import numpy as np
import pandas as pd
from pyfaidx import Fasta

from .classes import Feature, Seqinfo

logger = logging.getLogger('panfeed.input')


def what_are_my_inputfiles(gffdir, fastadir=None): # adds all the GFF files in the directory to a list
    filelist = set()
    fastalist = set()

    for file in os.listdir(gffdir):
        only_file = os.path.split(file)[-1]
        if file.endswith(".gff"):
            genome = ".".join(only_file.split(".")[:-1])
            logger.debug(f"Adding {genome} ({file})")
            filelist.add(genome)
        else:
            logger.debug(f"Ignoring {file}")

    if fastadir is not None:
        # check which of the GFF files have a corresponding
        # nucleotide fasta file
        for file in os.listdir(fastadir):
            only_file = os.path.split(file)[-1]
            if file.endswith(".fasta") or file.endswith(".fna"):
                genome = ".".join(only_file.split(".")[:-1])
                if genome in filelist:
                    logger.debug(f"Adding {genome} nucleotides ({file})")
                    fastalist.add(genome)
                else:
                    logger.debug(f"Ignoring {genome} (no corresponding gff)")
            else:
                logger.debug(f"Ignoring {file}")


    if len(filelist) == 0:
        logger.error(f"No GFF files found in working directory ({gffdir})")
        sys.exit(1)
        
    return sorted(filelist), sorted(fastalist)


def prep_data_n_fasta(filelist, fastalist,
                      gffdir, fastadir, output): # prepares the hybrid GFF files and creates fasta files
    data = {}
    
    for genome in filelist:
        logger.debug(f"Handling {genome}")

        if genome not in fastalist:
            logger.debug(f"Creating fasta file for {genome}")
            
            txtfile = open(os.path.join(output, f"{genome}.fasta"), "w")
            txtfile.write(open(os.path.join(gffdir, f"{genome}.gff"), "r").read().split("##FASTA")[1])
            txtfile.close()
        
            logger.debug(f"Creating faidx file for {genome}")
            sequences = create_faidx(os.path.join(output, f'{genome}.fasta'))
        else:
            logger.debug(f"Creating faidx file for existing nucleotide sequence {genome}")
            file1 = os.path.join(fastadir, f'{genome}.fna')
            file2 = os.path.join(fastadir, f'{genome}.fasta')
            if os.path.exists(file1):
                ffile = file1
            elif os.path.exists(file2):
                ffile = file2
            else:
                logger.error(f"Neither {genome}.fna not {genome}.fasta found in {fastadir}")
                sys.exit(1)
            sequences = create_faidx(ffile)

        logger.debug(f"Parsing features for {genome}")
        features = parse_gff(os.path.join(gffdir, f'{genome}.gff'))
    
        data[genome] = (sequences, features)
    
    return data


def clean_up_fasta(filelist, fastalist, output, fastadir):
    for genome in filelist:
        if genome in fastalist:
            logger.debug(f"Keeping fasta file for {genome}")
        else:
            fasta_file = os.path.join(output, f"{genome}.fasta")
            logger.debug(f"Removing fasta file for {genome} ({fasta_file})")
            if os.path.exists(fasta_file):
                os.remove(fasta_file)
            else:
                logger.warning(f"Could not delete {fasta_file}")
        
        if genome not in fastalist:
            faidx_file = os.path.join(output, f"{genome}.fasta.fai")
        else:
            faidx_file1 = os.path.join(fastadir, f"{genome}.fasta.fai")
            faidx_file2 = os.path.join(fastadir, f"{genome}.fna.fai")
            if os.path.isfile(faidx_file1):
                faidx_file = faidx_file1
            else:
                faidx_file = faidx_file2
        logger.debug(f"Removing faidx file for {genome} ({faidx_file})")
        if os.path.exists(faidx_file):
            os.remove(faidx_file)
        else:
            logger.warning(f"Could not delete {faidx_file}")


def set_input_output(stroi_in, genes_in, presence_absence, output,
                     single_file=True, compress=False):
    # determines the names of the input, output files 
   
    logger.debug(f"Loading pangenome file from panaroo ({presence_absence})")
    genepres = pd.read_csv(presence_absence,
                           sep=",", index_col=0,
                           low_memory=False).drop(
                                   columns=['Non-unique Gene name', 'Annotation'])
    # load the gene_presence_absence matrix which will be used to ascribe strain/genes to clusters
    
    if stroi_in is not None:
        logger.debug(f"Loading target strains ({stroi_in})")
        stroi = set()
        strainfile = open(f"{stroi_in}", "r")
        for line in strainfile:
            stroi.add(line.rstrip("\n"))
    else:
        logger.warning(f"No target strains provided")
        stroi = ""

    if genes_in is not None:
        logger.debug(f"Loading target gene clusters ({genes_in})")
        genes = set()
        genesfile = open(f"{genes_in}", "r")
        for line in genesfile:
            genes.add(line.rstrip("\n"))
    else:
        genes = None

    if os.path.exists(output):
        logger.error(f"Output directory {output} exists! Please "
                      "remove it and restart")
        sys.exit(1)
    else:
        logger.debug(f"Creating output directory ({output})")
        os.mkdir(output)
    
    logger.debug(f"Creating output files within {output}")
    
    if single_file:
        kmer_stroi = create_kmer_stroi(output, compress)
        hash_pat, kmer_hash = create_hash_files(output, compress) 
    else:
        # delayied opening of the files
        kmer_stroi = None
        hash_pat = None
        kmer_hash = None

    return stroi, genes, kmer_stroi, hash_pat, kmer_hash, genepres


def create_kmer_stroi(output, compress=False):
    if not compress:
        kmer_stroi = open(os.path.join(output, "kmers.tsv"), "w")
    else:
        kmer_stroi = gzip.open(os.path.join(output, "kmers.tsv.gz"), "wt",
                               compresslevel=9)
    
    #creates the header for the strains of interest output file
    kmer_stroi.write("cluster\tstrain\tfeature_id\tcontig\tfeature_strand\tcontig_start\tcontig_end\tgene_start\tgene_end\tstrand\tk-mer\n")
    kmer_stroi.flush()

    return kmer_stroi


def create_hash_files(output, compress=False):
    if not compress:
        hash_pat = open(os.path.join(output, "hashes_to_patterns.tsv"), "w")
        kmer_hash = open(os.path.join(output, "kmers_to_hashes.tsv"), "w")
    else:
        hash_pat = gzip.open(os.path.join(output, "hashes_to_patterns.tsv.gz"),
                             "wt", compresslevel=9)
        kmer_hash = gzip.open(os.path.join(output, "kmers_to_hashes.tsv.gz"),
                              "wt", compresslevel=9)

    return hash_pat, kmer_hash


def create_faidx(file_name, uppercase=True):
    
    faidx = Fasta(file_name,
                  sequence_always_upper=uppercase)
    
    return faidx


def parse_gff(file_name, feature_types=None):
    if feature_types is None:
        feature_types = {'CDS'}

    # output dict
    # key: feature ID
    # value: Feature NamedTuple
    features = {}
    
    with open(file_name, 'r') as gff:
        for line in gff:
            if line.lstrip().startswith('##FASTA'):
                # start of FASTA entries, end of file
                break
            
            elif line.lstrip().startswith('#'):
                # comment, ignore
                continue
            
            # should be a valid GFF3 line
            entries = line.split('\t')
            
            try:
                ftype = entries[2]
                
                if ftype not in feature_types:
                    continue

                chrom = entries[0]
                start = int(entries[3])
                end = int(entries[4])
                strand = entries[6]
                
                # integer takes up less space
                if strand == '+':
                    strand = 1
                else:
                    strand = -1

                # fetch the feature ID from the last field
                ID = None
                for entry in entries[8].split(';'):
                    if entry.startswith('ID') and '=' in entry:
                        ID = entry.split('=')[1]
                        
                # could not find it, skip this entry
                if ID is None:
                    continue

                # save the relevant details
                features[ID] = Feature(ID, chrom, start, end, strand)
                
            except Exception as e:
                # not distinguishing between exceptions
                # not great behaviour
                logger.warning(f'{e}, skipping line "{line.rstrip()}" from {file_name}')                
                continue
    
    return features


def iter_gene_clusters(panaroo, genome_data, up, down, down_start_codon, patfilt,
                       gene_list=None):
    
    # go through each gene cluster
    all_ogs = panaroo.shape[0]
    for i, (idx, row) in enumerate(panaroo.iterrows()):
        if gene_list is not None and idx not in gene_list:
            logger.debug(f"Skipping {idx} ({i+1}/{all_ogs})")
            continue

        logger.debug(f"Extracting sequences from {idx} ({i+1}/{all_ogs})")
        # output dict
        # key: strain
        # value: list of faidx sequence objects
        gene_sequences = {}
        
        # keep track of who has the gene
        # could be useful
        # to prepopulate the presence/absence vector
        strains = row.index
        
        sortstrain = {x: i
                      for i, x in enumerate(sorted(strains))}
        
        present = row.dropna().index
        
        absent = strains.difference(present)
            
        clusterpresab = np.zeros(len(strains), dtype = int)
        for strain in present:
            clusterpresab[sortstrain[strain]] = 1 

        # print(clusterpresab)
        # cycle through all the strains that have the gene
        for strain, genes in row.dropna().items():
            
            strain = str(strain)
            
            gene_sequences[strain] = []
            # access the GFF/fasta data
            sequences, features = genome_data[strain]
            # be aware of paralogs
            for gene in genes.split(';'):
                # access a particular gene
                try:
                    feat = features[gene]
                except KeyError:
                    # e.g. refound genes are not in the GFF
                    logger.warning(f"Could not find gene {gene} from {idx} in {strain}")
                    continue
                
                # corner case: upstream offset is over the contig's edge
                # adjust if so and save the true offset
                if feat.strand > 0 and feat.start-1-up < 0:
                    offset = feat.start - 1
                elif feat.strand < 0 and feat.end+up > len(sequences[feat.chromosome]):
                    offset = len(sequences[feat.chromosome]) - feat.end
                else:
                    offset = up
                
                # access its gene sequence
                if not down_start_codon:
                    # down means relative to the gene's 3'
                    if feat.strand > 0:
                        seq = sequences[feat.chromosome][feat.start-1-offset:feat.end+down]
                        seq_start = feat.start - offset
                        seq_end = feat.end + down
                    else:
                        seq = -sequences[feat.chromosome][feat.start-1-down:feat.end+offset]
                        seq_start = feat.start - down
                        seq_end = feat.end + offset
                else:
                    # down means relative to the gene's 5'
                    if feat.strand > 0:
                        seq = sequences[feat.chromosome][feat.start-1-offset:feat.start+down]
                        seq_start = feat.start - offset
                        seq_end = feat.start + down
                    else:
                        seq = -sequences[feat.chromosome][feat.end-1-down:feat.end+offset]
                        seq_start = feat.end - down
                        seq_end = feat.end + offset

                # reverse complement
                revseq = -seq
                # reverse it, easier to slice
                # so effectively now it's just the complement
                revseq = revseq[::-1]

                # save it
                seq1 = Seqinfo(str(seq), str(revseq),
                               feat.id, feat.chromosome,
                               seq_start, seq_end,
                               feat.strand, offset)
                gene_sequences[strain].append(seq1)
        
        # provide an empty list if the strain does not have the gene
        for strain in absent:
            gene_sequences[strain] = []
            
        yield gene_sequences, idx, clusterpresab


if __name__ == "__main__":
    pass



#!/usr/bin/env python


import sys
import logging
import argparse
import numpy as np
import pandas as pd

from .__init__ import __version__
from .colorlog import ColorFormatter

logger = logging.getLogger('panfeed')


def set_logging(v):
    logger.propagate = True
    logger.setLevel(logging.DEBUG)

    ch = logging.StreamHandler()
    if v == 0:
        ch.setLevel(logging.INFO)
    elif v >= 1:
        ch.setLevel(logging.DEBUG)
    formatter = ColorFormatter('%(asctime)s - %(name)s - $COLOR%(message)s$RESET','%H:%M:%S')
    ch.setFormatter(formatter)
    logger.addHandler(ch)


def get_options():
    description = "Annotate association results with positional information"

    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("-a", "--associations",
            required=True,
            help="TSV file containing hashes and their significance "
                 "(e.g. pyseer output; tab-delimited, should have a header"
                 ", first column contains "
                 "the hash, and another column - by default 'lrt-pvalue' - "
                 "contains the association p-value, "
                 "which can be used to filter the table)")

    parser.add_argument("-p", "--kmers-to-hashes",
            required=True,
            help="TSV file relating gene clusters, kmers, and their hashes "
                 "(i.e. panfeed's kmers_to_hases.tsv file)")

    parser.add_argument("-k", "--kmers",
            required=True,
            help="TSV file with positional information of individual k-mers "
                 "(i.e. panfeed's kmers.tsv file)")

    parser.add_argument("-t", "--threshold",
            type=float,
            default=1,
            help="Association p-value threshold (default %(default).2f)")

    parser.add_argument("-c", "--column",
            default='lrt-pvalue',
            help="P-value column in the associations file (default %(default)s)")

    parser.add_argument("-o", "--output",
            default=None,
            help="Filename to save filtered associations table (not saved by default)")

    parser.add_argument("--only-passing",
            action="store_true",
            default=False,
            help="Only output passing k-mers (default is all)")

    parser.add_argument("-v", action='count',
                        default=0,
                        help='Increase verbosity level')
    parser.add_argument('--version', action='version',
                        version='%(prog)s '+__version__)

    return parser.parse_args()


def main():
    args = get_options()

    set_logging(args.v)

    a = pd.read_csv(args.associations, sep='\t', index_col=0)
    a.index.name = 'hashed_pattern'
    if args.column not in a.columns:
        logger.warning(f"Associations file does not have the {args.column} column")
        sys.exit(1)

    # filter associations table
    a = a[a[args.column] <= args.threshold]
    passing_hashes = set(a.index)

    logger.info(f"{len(passing_hashes)} patterns pass the p-value threshold {args.threshold}")
    if args.output is not None:
        a.to_csv(args.output, sep='\t')
        logger.info(f"Saved filtered associations to {args.output}")

    # load kmers to hashes table, piece-wise to reduce memory footprint
    # thanks to SO: https://stackoverflow.com/a/13653490/1237531
    iter_h = pd.read_csv(args.kmers_to_hashes, sep='\t',
                         iterator=True, chunksize=100_000)
    h = pd.concat([x[x['hashed_pattern'].isin(passing_hashes)]
        for x in iter_h]).set_index('hashed_pattern')
    clusters = set(h['cluster'].unique())
    kmers = set(h['k-mer'].unique())

    logger.info(f"Found {len(clusters)} gene clusters")
    logger.info(f"Found {len(kmers)} k-mers")

    # load k-mers table piece-wise to reduce memory footprint
    # thanks to SO: https://stackoverflow.com/a/13653490/1237531
    iter_k = pd.read_csv(args.kmers, sep='\t',
                         iterator=True, chunksize=100_000)
    k = pd.concat([x[x['cluster'].isin(clusters)]
        for x in iter_k]).set_index(['cluster', 'k-mer'])

    a = a.join(h, how='inner')
    if args.only_passing:
        how = 'left'
    else:
        how = 'right'
    a = a.reset_index().set_index(['cluster', 'k-mer']).join(k, how=how)

    a.to_csv(sys.stdout, sep='\t')


if __name__ == "__main__":
    main()



#!/usr/bin/env python


import sys
import logging
import argparse
import logging.handlers
from functools import partial
import itertools
from multiprocessing import Process, Queue

from .__init__ import __version__
from .colorlog import ColorFormatter

from .input import prep_data_n_fasta, what_are_my_inputfiles, set_input_output
from .input import clean_up_fasta, what_are_my_inputfiles, set_input_output
from .panfeed import cluster_cutter, pattern_hasher, write_headers
from .input import iter_gene_clusters


logger = logging.getLogger('panfeed')


def set_logging(v):
    logger.propagate = True
    logger.setLevel(logging.DEBUG)

    ch = logging.StreamHandler()
    if v == 0:
        ch.setLevel(logging.INFO)
    elif v >= 1:
        ch.setLevel(logging.DEBUG)
    formatter = ColorFormatter('%(asctime)s - %(name)s - $COLOR%(message)s$RESET','%H:%M:%S')
    ch.setFormatter(formatter)
    logger.addHandler(ch)


def worker(f, read_q, write_q):
    while True:
        work = read_q.get()
        if work is None:
            logger.debug("Worker process finished")
            write_q.put(None)
            return

        ret = f(work)
        if len(ret) == 0:
            logger.debug("Worker process finished")
            write_q.put(None)
            return
        write_q.put((ret,),timeout = 300)


def reader(iter_i, read_q, n):
    for x in iter_i:
        read_q.put(x)

    logger.debug("Reader process finished")

    # poison pill
    # one for each worker
    for _ in range(n):
        read_q.put(None)


def writer(f, write_q, n):
    dead = 0

    patterns = set()

    while True:
        work = write_q.get()
        if work is None:
            dead += 1
            logger.debug(f"Finished worker processes {dead}/{n}")
            if dead == n:
                logger.debug("All worker processes are finished")
                return
            continue
        patterns = f(work, patterns=patterns)


def get_options():
    description = "Get gene cluster specific k-mers from a set of bacterial genomes"
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("-g", "--gff",
                        required=True,
                        help = "Directory containing all samples' GFF "
                               "files (must contain nucleotide sequence as "
                               "well unless -f is used, "
                               "and samples should be named in the "
                               "same way as in the panaroo header)")

    parser.add_argument("-p", "--presence-absence",
                        required=True,
                        help = "Gene clusters presence absence table "
                               "as output by panaroo")

    parser.add_argument("--targets",
                        default=None,
                        help = "File indicating for which samples "
                               "k-mer positions should be logged "
                               "(one sample name per line, "
                               "default: no samples)")

    parser.add_argument("--genes",
                        default=None,
                        help = "File indicating for which gene clusters "
                               "k-mer positions and patterns should be logged "
                               "(one gene cluster ID per line, "
                               "default: all gene clusters)")

    parser.add_argument("-o", "--output",
                        default = "panfeed",
                        help = "Output directory to store outputs "
                               "(will cause an error if already present)")

    parser.add_argument("-f", "--fasta",
                        help = "Directory containing all samples' nucleotide "
                               "fasta files (extension either .fasta "
                               "or .fna, "
                               "samples should be named in the "
                               "same way as in the panaroo header")

    parser.add_argument("-k", "--kmer-length", type = int,
                        default = 31,
                        help = "K-mer length (default: %(default)d)")

    parser.add_argument("--maf", type = float,
                        default = 0.01,
                        help = "Minor allele frequency threshold; "
                               "patterns whose frequency is below "
                               "this value or above 1-MAF are excluded "
                               "(default: %(default).2f, does not apply "
                               "to the kmers.tsv file)")

    parser.add_argument("--upstream", type = int,
                        default = 0,
                        help = "How many bases to include upstream of "
                               "the actual gene sequences "
                               "(e.g. to include the 5' region, "
                               "default: %(default)d)")

    parser.add_argument("--downstream", type = int,
                        default = 0,
                        help = "How many bases to include downstream of "
                               "the actual gene sequences "
                               "(e.g. to include the 3' region, "
                               "default: %(default)d)")

    parser.add_argument("--downstream-start-codon",
                        action = "store_true",
                        default = False,
                        help = "Center the --downstream argument "
                               "at the start codon (default is stop codon)")

    parser.add_argument("--non-canonical",
                        action = "store_true",
                        default = False,
                        help = "Compute non-canonical k-mers "
                               "(default is canonical)")

    parser.add_argument("--no-filter",
                        action = "store_true",
                        default = False,
                        help = "Do not filter out k-mers with the same "
                               "presence absence pattern as the gene "
                               "cluster itself")

    parser.add_argument("--consider-missing",
                        action = "store_true",
                        default = False,
                        help = "Output NaN for strains that do not encode "
                               "for a k-mer if the gene is missing "
                               "(default: value is set to 0, as for the gene. "
                               "WARNING: slows down execution)")

    parser.add_argument("--multiple-files",
                        action = "store_true",
                        default = False,
                        help = "Generate one set of outputs for each "
                               "gene cluster (default: one set of outputs)")

    parser.add_argument("--compress",
                        action = "store_true",
                        default = False,
                        help = "Compress output files with gzip "
                               "(default: plain text)")

    parser.add_argument("--cores",
                        type = int,
                        default = 1,
                        help = "Threads (default: %(default)d, at least 3 "
                               "are needed for parallelization)")

    parser.add_argument("-ql", "--queue-limit",
                        type = int,
                        default = 3,
                        help = "limit on items that may be put into the reading"
                               "and writing queues. (default: %(default)d)"
                               "this option is only relevant for cores > 1"
                               "reading queue limit = ql * cores"
                               "writing queue limit = ql")

    parser.add_argument("-v", action='count',
                        default=0,
                        help='Increase verbosity level')
    parser.add_argument('--version', action='version',
                        version='%(prog)s '+__version__)

    return parser.parse_args()


def main():
    args = get_options()

    set_logging(args.v)

    qlimit = args.queue_limit

    klength = args.kmer_length

    if args.downstream_start_codon == True and args.upstream + args.downstream < klength:
        logger.warning("Query sequence is shorter than k-mer length"
                       "Decrease k-mer size or increase query sequence length")
        sys.exit(1)

    if args.maf > 0.5:
        logger.warning("--maf should be below 0.5")
        sys.exit(1)

    logger.info("Looking at input GFF files")
    filelist, fastalist = what_are_my_inputfiles(args.gff, args.fasta)
    logger.info(f"Found {len(filelist)} input genomes")
    if args.fasta is not None:
        logger.info(f"Found {len(fastalist)} input nucleotide sequences")

    data = {}

    logger.info("Preparing output files")
    (stroi, genes, kmer_stroi, hash_pat,
     kmer_hash, genepres) = set_input_output(args.targets,
                                             args.genes,
                                             args.presence_absence,
                                             args.output,
                                             not args.multiple_files,
                                             args.compress)

    logger.info("Preparing inputs")
    data = prep_data_n_fasta(filelist, fastalist,
                             args.gff, args.fasta, args.output)

    if not args.multiple_files:
        write_headers(hash_pat, kmer_hash, genepres)

    logger.info("Extracting k-mers")
    iter_i = iter_gene_clusters(genepres, 
                                data,
                                args.upstream,
                                args.downstream,
                                args.downstream_start_codon,
                                not args.no_filter,
                                genes)
    iter_o = partial(cluster_cutter,
                     klength=klength,
                     stroi=stroi,
                     multiple_files=args.multiple_files,
                     canon=not args.non_canonical,
                     consider_missing_cluster=args.consider_missing,
                     output=args.output,
                     compress=args.compress)

    patterns = set()
    func_w = partial(pattern_hasher,
                     kmer_stroi=kmer_stroi,
                     hash_pat=hash_pat,
                     kmer_hash=kmer_hash,
                     genepres=genepres,
                     patfilt=not args.no_filter,
                     maf=args.maf,
                     consider_missing_cluster=args.consider_missing,
                     output=args.output,
                     patterns=patterns,
                     compress=args.compress)

    if args.cores > 2:
        # thanks to @SamStudio8 for the inspiration
        read_q = Queue(maxsize = qlimit * args.cores)
        write_q = Queue(maxsize = qlimit)

        processes = []

        reader_process = Process(
            target=reader,
            args=(
                iter_i,
                read_q,
                args.cores - 2
            ),
        )
        processes.append(reader_process)

        writer_process = Process(
            target=writer,
            args=(
                func_w,
                write_q,
                args.cores - 2,
            ),
        )
        processes.append(writer_process)

        for i in range(args.cores - 2):
            p = Process(
                target=worker,
                args=(
                    iter_o,
                    read_q,
                    write_q,
                ),
            )
            processes.append(p)
        logger.debug(f"Started {args.cores - 2} worker processes")

        for p in processes:
            p.start()

        for p in processes:
            p.join()
    else:
        if args.cores > 1:
            logger.warning("Need at least 3 cores for parallelization")
            logger.warning("Running a single thread")

        patterns = set()
        for x in iter_i:
            ret = iter_o(x)
            if len(ret) == 0:
                continue
            patterns = func_w((ret,),
                              patterns=patterns)


    if kmer_stroi is not None:
        kmer_stroi.close()

    if hash_pat is not None:
        hash_pat.close()

    if kmer_hash is not None:
        kmer_hash.close()

    logger.info("Removing temporary fasta files and faidx indices")
    clean_up_fasta(filelist, fastalist, args.output, args.fasta)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import os
import logging
import numpy as np
import hashlib
import binascii
from copy import deepcopy
from io import StringIO

from .input import create_hash_files, create_kmer_stroi

logger = logging.getLogger('panfeed.panfeed')


def init_presabs_vector(n_strains, clusterpresab, missing_nan=False):
    v = np.zeros(n_strains, dtype=np.float64)
    if missing_nan:
        v[clusterpresab == 0] = np.nan
    return v


def cluster_cutter(cluster_gen, klength, stroi,
                   multiple_files, canon, consider_missing_cluster,
                   output, compress=False):
#iterates through genes present in the cluster
#shreds gene sequence into k-mers and adds them to the cluster dictionary
#if a gene belongs to a strain of interest, additional info on the k-mer is saved
    kmer_stroi = None

    cluster, idx, clusterpresab = cluster_gen
    logger.debug(f"Extracting k-mers from {idx}")

    memchunk = StringIO()
    if multiple_files:
        # close existing file?
        if kmer_stroi is not None:
            kmer_stroi.close()
        # we need to create the file
        path = os.path.join(output, idx)
        if not os.path.exists(path):
            os.mkdir(path)
        kmer_stroi = create_kmer_stroi(path, compress)

    cluster_dict = {}

    sortstrain = {x: i
                    for i, x in enumerate(sorted(cluster.keys()))}
    n_strains = len(sortstrain)

    base_vector = init_presabs_vector(n_strains, clusterpresab,
                                      consider_missing_cluster)

    for strain in cluster.keys():
        for seq in cluster[strain]:
            gene_id = seq.id
            offset = seq.offset

            num_kmer = len(seq.sequence) - klength + 1
            seqlen = len(seq.sequence)

            contig = seq.chromosome
            strand = seq.strand
            for pos in range(num_kmer):
                specseq = str(seq.sequence[pos:pos + klength])
                # get back to reverse complement
                revspecseq = str(seq.compsequence[pos:pos + klength])[::-1]
            
                if canon == True:
                    if specseq <= revspecseq:
                        canonseq = specseq
                        used_strand = 1
                    else:
                        canonseq = revspecseq
                        used_strand = -1 

                    cluster_dict[canonseq] = cluster_dict.get(canonseq,
                                                              deepcopy(base_vector))
                    cluster_dict[canonseq][sortstrain[strain]] = 1
                else:
                    used_strand = strand
                    cluster_dict[specseq] = cluster_dict.get(specseq,
                                                             deepcopy(base_vector))
                    cluster_dict[specseq][sortstrain[strain]] = 1

                    cluster_dict[revspecseq] = cluster_dict.get(revspecseq,
                                                                deepcopy(base_vector))
                    cluster_dict[revspecseq][sortstrain[strain]] = 1

                if strain in stroi:
                    if strand > 0:
                        contigstart = seq.start
                        truestart = contigstart + pos
                        trueend = contigstart + pos + klength

                    else:
                        contigstart = seq.end
                        trueend = contigstart - pos
                        truestart = contigstart - pos - klength

                    genestart = pos - offset
                    geneend = pos + klength - offset
                    if canon == True:
                        memchunk.write(f"{idx}\t{strain}\t{gene_id}\t{contig}\t{strand}\t{truestart}\t{trueend}\t{genestart}\t{geneend}\t{used_strand}\t{canonseq}\n")
                    else:
                        memchunk.write(f"{idx}\t{strain}\t{gene_id}\t{contig}\t{strand}\t{truestart}\t{trueend}\t{genestart}\t{geneend}\t{used_strand}\t{specseq}\n")
                        memchunk.write(f"{idx}\t{strain}\t{gene_id}\t{contig}\t{strand}\t{truestart}\t{trueend}\t{genestart}\t{geneend}\t{-used_strand}\t{revspecseq}\n")

    if multiple_files:
        kmer_stroi.write(memchunk.getvalue())
        return idx, cluster_dict, clusterpresab, None
    else:
        return idx, cluster_dict, clusterpresab, str(memchunk.getvalue())
 

def write_headers(hash_pat, kmer_hash, genepres):
    memchonkheader1 = StringIO()
    memchonkheader2 = StringIO()
    
    memchonkheader1.write("hashed_pattern")
    for strain in sorted(genepres.columns):
        memchonkheader1.write(f"\t{strain}")
    memchonkheader1.write("\n")
    hash_pat.write(memchonkheader1.getvalue())
    hash_pat.flush()

    memchonkheader2.write("cluster\tk-mer\thashed_pattern\n")
    kmer_hash.write(memchonkheader2.getvalue())
    kmer_hash.flush()


def pattern_hasher(cluster_dict_iter, kmer_stroi, hash_pat, kmer_hash,
                   genepres, patfilt, maf, output, patterns=None,
                   consider_missing_cluster=False,
                   compress=False):
    #iterates through the cluster dictionary output by cluster_cutter()
    #outputs the hashed patterns, patterns and k-mers to the output files
    #two files are created: hashed k-mer patterns to presence/absence patterns
    #                       k-mer sequences to hashed k-mer patterns
    
    # set flag for multiple files
    multiple_files = False
    if hash_pat is None or kmer_hash is None:
        multiple_files = True
    
    # keep track of already observed patterns
    # might have a big memory footprint
    # TODO: check for memory footprint
    if patterns is None:
        patterns = set()

    for idx, cluster_dict, clusterpresab, memchunk in cluster_dict_iter:
        if multiple_files:
            # close existing file?
            if hash_pat is not None:
                hash_pat.close()
            if kmer_hash is not None:
                kmer_hash.close()
            # we need to create the file
            path = os.path.join(output, idx)
            if not os.path.exists(path):
                os.mkdir(path)
            hash_pat, kmer_hash = create_hash_files(path, compress)
            # must "flush" the already observed patterns file
            patterns = set()

            write_headers(hash_pat, kmer_hash, genepres)
        else:
            if memchunk is not None:
                kmer_stroi.write(memchunk)

        memchunkhash_pat = StringIO()
        memchunkkmer_hash = StringIO()
       
        hashed = hashlib.md5(clusterpresab.view(np.uint8))
        khash = binascii.b2a_base64(hashed.digest()).decode()[:24]
        memchunkkmer_hash.write(f"{idx}\t\t{khash}\n")
        
        if khash not in patterns:
            patterns.add(khash)
            if not consider_missing_cluster:
                patterntup = "\t".join(map(str, clusterpresab.astype(np.uint8)))
            else:
                patterntup = "\t".join(['' if np.isnan(x)
                                        else str(int(x))
                                        for x in clusterpresab])
            memchunkhash_pat.write(f"{khash}\t{patterntup}\n")
        
        for kmer in cluster_dict:
            if not consider_missing_cluster:
                af = cluster_dict[kmer].sum() / cluster_dict[kmer].shape[0]
            else:
                # be aware of missing values
                k_presab = cluster_dict[kmer]
                k_presab = k_presab[~np.isnan(k_presab)]
                af = k_presab.sum() / k_presab.shape[0]
            if af >= 0.5:
                af = 1-af
            if af < maf:
                continue
            
            if patfilt == True:
                if tuple(cluster_dict[kmer]) == tuple(clusterpresab):
                    continue

            hashed = hashlib.md5(cluster_dict[kmer].view(np.uint8))
            khash = binascii.b2a_base64(hashed.digest()).decode()[:24]
            memchunkkmer_hash.write(f"{idx}\t{kmer}\t{khash}\n")
            
            if khash in patterns:
                continue
            patterns.add(khash)

            if not multiple_files and not len(patterns) % 1000:
                logger.debug(f"Observed patterns so far: {len(patterns)}")

            if not consider_missing_cluster:
                patterntup = "\t".join(map(str, cluster_dict[kmer].astype(np.uint8)))
            else: 
                patterntup = "\t".join(['' if np.isnan(x)
                                        else str(int(x))
                                        for x in cluster_dict[kmer]])
            memchunkhash_pat.write(f"{khash}\t{patterntup}\n")

        hash_pat.write(memchunkhash_pat.getvalue())
        kmer_hash.write(memchunkkmer_hash.getvalue())

        if multiple_files:
            memchunkhash_pat = None
            memchunkkmer_hash = None

    hash_pat.flush()
    kmer_hash.flush()

    return patterns

if __name__ == "__main__":
    pass



#!/usr/bin/env python


import sys
import logging
import argparse
import numpy as np
import pandas as pd

from matplotlib import colors
import matplotlib.pyplot as plt
import seaborn as sns
from mpl_toolkits.axes_grid1 import make_axes_locatable

from .__init__ import __version__
from .colorlog import ColorFormatter

logger = logging.getLogger('panfeed')


def set_logging(v):
    logger.propagate = True
    logger.setLevel(logging.DEBUG)

    ch = logging.StreamHandler()
    if v == 0:
        ch.setLevel(logging.INFO)
    elif v >= 1:
        ch.setLevel(logging.DEBUG)
    formatter = ColorFormatter('%(asctime)s - %(name)s - $COLOR%(message)s$RESET','%H:%M:%S')
    ch.setFormatter(formatter)
    logger.addHandler(ch)


def get_options():
    description = "Plot association results from panfeed"

    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("-k", "--kmers",
            required=True,
            help="TSV file containing the output of panfeed-get-kmers")

    parser.add_argument("-c", "--column",
            default="lrt-pvalue",
            help="P-value column in the associations file (default %(default)s)")

    parser.add_argument("-t", "--threshold",
            type=float,
            default=1,
            help="Association p-value threshold (default %(default).2f)")

    parser.add_argument("-p", "--phenotype",
            required=True,
            help="Phenotype file in TSV format, used to list all strains")

    parser.add_argument("--phenotype-column",
            default=None,
            help="Column in phenotype TSV file for sorting "
                 "(default is sorting by p-values)")

    parser.add_argument('--sample',
                        type=float,
                        default=None,
                        help="Only show a randomly picked set of sample "
                             "(a value between 0 and 1 indicating the proportion"
                             " to show, default all)")

    parser.add_argument('--start',
                        type=int,
                        default=None,
                        help="Relative position to start the plots "
                             "(default all available positions)")

    parser.add_argument('--stop',
                        type=int,
                        default=None,
                        help="Relative position to end the plots "
                             "(default all available positions)")

    parser.add_argument('--format',
                        choices=('png',
                                 'tiff',
                                 'pdf',
                                 'svg'),
                        default='png',
                        help='Output format for plots (default %(default)s)')

    parser.add_argument('--dpi',
                        type=int,
                        default=300,
                        help='Output resolution (DPI, default %(default)d)')

    parser.add_argument("--minimum-pvalue",
            type=float,
            default=1E-10,
            help="Minimum p-value for color and transparency (default %(default).2e)")

    parser.add_argument("--nucleotides",
            action="store_true",
            default=False,
            help="Draw nucleotide sequence on all plots "
                 "(WARNING: only makes sense if few samples and positions are considered)")

    parser.add_argument('--alpha',
                        type=float,
                        default=0,
                        help="Opacity for non-passing k-mers "
                             "(between 0 and 1, 0 indicates full transparency"
                             ", default %(default).2f)")

    parser.add_argument('--xticks',
                        type=int,
                        default=200,
                        help="Spacing for ticks on x axis (default %(default)d)")

    parser.add_argument('--height',
                        type=int,
                        default=9,
                        help="Figure height (inches, default %(default)d)")

    parser.add_argument('--width',
                        type=int,
                        default=10,
                        help="Figure width (inches, default %(default)d)")

    parser.add_argument("-v", action='count',
                        default=0,
                        help='Increase verbosity level')
    parser.add_argument('--version', action='version',
                        version='%(prog)s '+__version__)

    return parser.parse_args()


def main():
    args = get_options()

    set_logging(args.v)

    if args.sample is not None and (args.sample > 1 or args.sample < 0):
        logger.warning("--sample should be between 0 and 1")
        sys.exit(1)
    if args.alpha > 1 or args.alpha < 0:
        logger.warning("--alpha should be between 0 and 1")
        sys.exit(1)
    if (args.start is not  None and args.stop is None) or (args.start is None and args.stop is not None):
        logger.warning("both --start and --stop are needed")
        sys.exit(1)
    if args.start is not None and args.start > args.stop:
        logger.warning("--start should be lower than --stop")
        sys.exit(1)
    if args.nucleotides and (args.sample is None or args.start is None):
        logger.warning("drawing nucleotide sequences without zooming in might "
                       "increase plotting time and memory consumption "
                       "while generating useless plots")

    p = pd.read_csv(args.phenotype, sep='\t', index_col=0)
    pbinary = False
    if args.phenotype_column is not None:
        if args.phenotype_column not in p.columns:
            logger.warning(f"phenotype file does not have the {args.phenotype_column} column")
            sys.exit(1)
        p = p[args.phenotype_column].dropna().sort_values(ascending=False)
        if args.sample is not None:
            p = p.sample(frac=args.sample)
        # guess is binary or continuous
        pvalues = set(p.values)
        if len(pvalues) == 2 and 1 in pvalues and 0  in pvalues:
            logger.info("Phenotype is binary")
            pbinary = True
            # index for transition from 1 to 0
            yindex = [i for i, x in enumerate(p)
                      if i > 0
                      and x != p.iloc[i - 1]]
            if len(yindex) > 0:
                yindex = yindex[0]
            else:
                yindex = None
        else:
            logger.info("Phenotype is continuos")
        logger.info(f"sorting samples by their {args.phenotype_column} phenotype")
    else:
        logger.info(f"sorting samples by their lowest association p-value")
        if args.sample is not None:
            p = p.sample(frac=args.sample)
    strains = set(p.index)
    logger.info(f"plots will include {len(strains)} samples")

    k = pd.read_csv(args.kmers, sep='\t')
    if args.column not in k.columns:
        logger.warning(f"k-mer file does not have the {args.column} column")
        sys.exit(1)
    # ignore samples not in phenotype file
    k = k[k['strain'].isin(strains)]
    # compute -log10 of p-value
    k['significance'] = -np.log10(k[args.column])
    clusters = set(k['cluster'].unique())

    bases = {'A': 'T',
             'T': 'A',
             'G': 'C',
             'C': 'G'}
    base2int = {'A': 0,
                'G': 1,
                'T': 2,
                'C': 3}

    # add a column with the nucleotide
    k['base'] = np.nan
    k['base'] = k['k-mer'].str.upper().str[0]
    k.loc[k[k['strand'] == -1].index,
          'base'] = [bases.get(x[-1], 'N')
                     for x in k[k['strand'] == -1
                         ]['k-mer'].str.upper().values]
    # convert it to scalar
    k['scalar'] = [base2int.get(x, np.nan) for x in k['base'].values]

    logger.info(f"preparing plots for {len(clusters)} gene clusters")

    def handle_paralogs(x):
        if len(x) > 1:
            return 99
        return x

    def handle_paralogs_text(x):
        if len(x) > 1:
            return '-'
        return x

    cmap1 = plt.get_cmap('viridis').copy()
    cmap1.set_bad('xkcd:grey')
    cmap1.set_under('xkcd:light grey')

    base_colors = list(sns.color_palette('tab20', 4))
    cmap2 = colors.LinearSegmentedColormap.from_list('nucleotides', base_colors, 4)
    cmap2.set_bad('xkcd:grey')
    cmap2.set_over(list(sns.color_palette('tab20', 5))[-1])

    for gene in sorted(clusters):
        logger.info(f"preparing plots for {gene}")

        if args.start is not None:
            # zoom in
            cl = k[(k['cluster'] == gene) &
                   (k['gene_start'] >= args.start) &
                   (k['gene_start'] <= args.stop)]
        else:
            cl = k[k['cluster'] == gene]
        # reality check
        if cl.shape[0] == 0:
            logger.warning(f"Skipping {gene}")
            continue

        # p-value matrix
        g = cl.pivot_table(index='strain', columns='gene_start',
                           values='significance',
                           aggfunc=max)
        # add missing strains for which no entry is present in k-mers table
        if len(strains.difference(g.index)) > 0:
            g = g.reindex(sorted(g.index) + sorted(strains.difference(g.index)))
        if args.phenotype_column is None:
            # sort the table by p-values
            g = g.loc[g.fillna(0).T.max().sort_values(ascending=False).index]
        else:
            # sort by phenotype
            g = g.loc[p.index]

        # nucleotide matrix
        b = cl.pivot_table(index='strain', columns='gene_start',
                           values='scalar',
                           aggfunc=handle_paralogs)
        # add missing strains for which no entry is present in k-mers table
        if len(strains.difference(b.index)) > 0:
            b = b.reindex(sorted(b.index) + sorted(strains.difference(b.index)))
        if args.phenotype_column is None:
            # sort the table by p-values
            b = b.loc[g.index]
        else:
            # sort by phenotype
            b = b.loc[p.index]
        if args.nucleotides:
            # nucleotide matrix
            t = cl.pivot_table(index='strain', columns='gene_start',
                               values='base',
                               aggfunc=handle_paralogs_text)
            # add missing strains for which no entry is present in k-mers table
            if len(strains.difference(t.index)) > 0:
                t = t.reindex(sorted(t.index) + sorted(strains.difference(t.index)))
            if args.phenotype_column is None:
                # sort the table by p-values
                t = t.loc[g.index]
            else:
                # sort by phenotype
                t = t.loc[p.index]

        # find the index for the gene start (if present)
        gene_start = [i for i, x in enumerate(g.columns)
                      if x == 0]
        if len(gene_start) > 0:
            gene_start = gene_start[0]
        else:
            gene_start = None
        # create an index to draw xticks as relative gene positions
        xticks = [(i, x) for i, x in enumerate(g.columns)
                  if not x % args.xticks]

        fig, ax = plt.subplots(figsize=(args.width, args.height))
        im = ax.imshow(g.values, cmap=cmap1, vmin=-np.log10(args.threshold),
                       vmax=-np.log10(args.minimum_pvalue),
                       aspect="auto", interpolation='none',
                       rasterized=True, alpha=1)
        if args.nucleotides:
            for x in range(t.shape[1]):
                for y in range(t.shape[0]):
                    ax.text(x, y, t.iloc[y, x],
                            ha='center', va='center')
        ax.set_yticks([])
        if pbinary and yindex is not None:
            ax.axhline(yindex, lw=1, color="black")
        if gene_start is not None:
            ax.axvline(gene_start, lw=1, color="black")
        ax.set_xticks([x[0] for x in xticks],
                      labels=[x[1] for x in xticks])
        ax.set_title(f"significant k-mers {gene}")
        ax.set_xlabel("position relative to gene start")
        ax.set_ylabel(f"{g.shape[0]} samples")
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="2.5%", pad=0.05)
        coba = plt.colorbar(im, cax=cax)
        coba.set_label("-log10 p-value")
        plt.savefig(f'significance_{gene}.{args.format}', dpi=args.dpi, bbox_inches="tight")
        logger.info(f'saved plot significance_{gene}.{args.format}')
        plt.close()

        fig, ax = plt.subplots(figsize=(args.width, args.height))
        im = ax.imshow(b.values, cmap=cmap2, vmin=0, vmax=3,
                       aspect="auto", interpolation='none',
                       rasterized=True, alpha=1)
        if args.nucleotides:
            for x in range(t.shape[1]):
                for y in range(t.shape[0]):
                    ax.text(x, y, t.iloc[y, x],
                            ha='center', va='center')
        ax.set_yticks([])
        if pbinary and yindex is not None:
            ax.axhline(yindex, lw=1, color="black")
        if gene_start is not None:
            ax.axvline(gene_start, lw=1, color="black")
        ax.set_xticks([x[0] for x in xticks],
                      labels=[x[1] for x in xticks])
        ax.set_title(f"nucleotide sequence {gene}")
        ax.set_xlabel("position relative to gene start")
        ax.set_ylabel(f"{g.shape[0]} samples")
        plt.savefig(f'sequence_{gene}.{args.format}', dpi=args.dpi, bbox_inches="tight")
        logger.info(f'saved plot sequence_{gene}.{args.format}')
        plt.close()

        g = g.fillna(g.min().min())
        g = (g - -np.log10(args.threshold)) / (-np.log10(args.minimum_pvalue) - -np.log10(args.threshold))
        g[g < args.alpha] = args.alpha
        g[g > 1] = 1
        g[np.isnan(g)] = args.alpha

        fig, ax = plt.subplots(figsize=(args.width, args.height))
        im = ax.imshow(b.values, cmap=cmap2, vmin=0, vmax=3,
                       aspect="auto", interpolation='none',
                       rasterized=True, alpha=g.values)
        if args.nucleotides:
            for x in range(t.shape[1]):
                for y in range(t.shape[0]):
                    ax.text(x, y, t.iloc[y, x],
                            ha='center', va='center')
        ax.set_yticks([])
        if pbinary and yindex is not None:
            ax.axhline(yindex, lw=1, color="black")
        if gene_start is not None:
            ax.axvline(gene_start, lw=1, color="black")
        ax.set_xticks([x[0] for x in xticks],
                      labels=[x[1] for x in xticks])
        ax.set_title(f"significant k-mers {gene}")
        ax.set_xlabel("position relative to gene start")
        ax.set_ylabel(f"{g.shape[0]} samples")
        plt.savefig(f'hybrid_{gene}.{args.format}', dpi=args.dpi, bbox_inches="tight")
        logger.info(f'saved plot hybrid_{gene}.{args.format}')
        plt.close()

    # plot nucleotide color scheme
    plt.figure(figsize=(3, 0.7))

    sns.heatmap(data=[[0, 1, 2, 3]],
                annot=[['A', 'G', 'T', 'C']],
                fmt='s',
                annot_kws={'size': 16, 'weight': 'bold'},
                cmap=cmap2,
                cbar=False,
                yticklabels=False,
                xticklabels=False,
                square=True)

    plt.savefig(f'sequence_legend.{args.format}', dpi=args.dpi, bbox_inches="tight")
    logger.info(f"Saved nucleotide color key at sequence_legend.{args.format}")
    plt.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python


import sys
import logging
import argparse
import numpy as np
import pandas as pd

from .__init__ import __version__
from .colorlog import ColorFormatter

logger = logging.getLogger('panfeed')


def set_logging(v):
    logger.propagate = True
    logger.setLevel(logging.DEBUG)

    ch = logging.StreamHandler()
    if v == 0:
        ch.setLevel(logging.INFO)
    elif v >= 1:
        ch.setLevel(logging.DEBUG)
    formatter = ColorFormatter('%(asctime)s - %(name)s - $COLOR%(message)s$RESET','%H:%M:%S')
    ch.setFormatter(formatter)
    logger.addHandler(ch)


def get_options():
    description = "Indicate which genes clusters have significantly associated patterns"

    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("-a", "--associations",
            required=True,
            help="TSV file containing hashes and their significance "
                 "(e.g. pyseer output; tab-delimited, should have a header"
                 ", first column contains "
                 "the hash, and another column - by default 'lrt-pvalue' - "
                 "contains the association p-value, "
                 "which can be used to filter the table)")

    parser.add_argument("-p", "--kmers-to-hashes",
            required=True,
            help="TSV file relating gene clusters, kmers, and their hashes "
                 "(i.e. panfeed's kmers_to_hases.tsv file)")

    parser.add_argument("-t", "--threshold",
            type=float,
            default=1,
            help="Association p-value threshold (default %(default).2f)")

    parser.add_argument("-c", "--column",
            default='lrt-pvalue',
            help="P-value column in the associations file (default %(default)s)")

    parser.add_argument("-o", "--output",
            default=None,
            help="Filename to save filtered associations table (not saved by default)")

    parser.add_argument("-v", action='count',
                        default=0,
                        help='Increase verbosity level')
    parser.add_argument('--version', action='version',
                        version='%(prog)s '+__version__)

    return parser.parse_args()


def main():
    args = get_options()

    set_logging(args.v)

    a = pd.read_csv(args.associations, sep='\t', index_col=0)
    if args.column not in a.columns:
        logger.warning(f"Associations file does not have the {args.column} column")
        sys.exit(1)

    # filter associations table
    a = a[a[args.column] <= args.threshold]
    passing_hashes = set(a.index)

    logger.info(f"{len(passing_hashes)} patterns pass the association threshold")
    if args.output is not None:
        a.to_csv(args.output, sep='\t')
        logger.info(f"Saved filtered associations to {args.output}")

    # load kmers to hashes table, piece-wise to reduce memory footprint
    # thanks to SO: https://stackoverflow.com/a/13653490/1237531
    iter_h = pd.read_csv(args.kmers_to_hashes, sep='\t',
                         iterator=True, chunksize=100_000)
    h = pd.concat([x[x['hashed_pattern'].isin(passing_hashes)]
        for x in iter_h])
    clusters = set(h['cluster'].unique())

    logger.info(f"Found significant associations for {len(clusters)} gene clusters")

    for c in clusters:
        print(c)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

from collections import namedtuple

Feature = namedtuple('Feature', ['id', 
                                 'chromosome',
                                 'start', 
                                 'end', 
                                 'strand'])

Seqinfo = namedtuple("Seqinfo", ["sequence",
                                 "compsequence",
                                 "id",
                                 "chromosome",
                                 "start",
                                 "end",
                                 "strand",
                                 "offset"])

Sequence = namedtuple("Sequence", [])


if __name__ == "__main__":
    pass